package cashmachine.api;

import cashmachine.model.Account;

/**
 * The connection to the internal bank network.
 *
 * Although concurrent withdraws may happen, implementing classes are not bound
 * to synchronize access to their methods.
 */
public interface AccountService
{
	/**
	 * Returns true if the given {@code accountId} is a valid account identifier
	 * in the bank system and the represented account hasn't been banned before.
	 */
	boolean isValidAccountId(int accountId);

	/**
	 * Checks if the given {@code pin} matches the actual pin for the given
	 * {@code accountId} and if successful returns the account information. The
	 * pin of the returned account will always be set to 0.
	 *
	 * @returns user account information or null if {@code pin} does not match.
	 */
	Account validatePin(int accountId, int pin);

	/**
	 * Bans the given account from further withdrawal actions. Using a banned
	 * account id in any other method of this interface will always result in
	 * the error behaviour.
	 *
	 * @return true if the account has been banned or is banned now, false if
	 *         there is no account for the given {@code accountI}
	 */
	boolean banAccount(int accountId);

	/**
	 * Checks if the given {@code pin} matches the actual pin for the given
	 * {@code accountId} and if this account is authorized to withdraw from the
	 * given {@code withdrawalAccountId}. If successful, it returns the
	 * withdrawal account information, otherwise null. The pin of the returned
	 * account will always be set to 0.
	 *
	 * @returns the withdrawal account or null
	 */
	Account getWithdrawalAccount(int accountId, int pin, int withdrawalAccountId);

	/**
	 * Withdraws {@code amount} money from the given account. See
	 * {@link #getWithdrawalAccount(int, int, int)} for the initial checks this
	 * method will do before withdrawing. Furthermore, the given {@code amount}
	 * is checked against the withdrawal account's balance and limit.
	 *
	 * @returns returns a transaction number greater than 0 if checks against
	 *          balance and limit are ok and the money has been withdrawn
	 *          otherwise 0
	 * @throws CashMachineException
	 *             if the initial checks do not hold or amount isn't greater
	 *             than 0
	 */
	int withdraw(int accountId, int pin, int withdrawalAccountId, int amount)
	throws CashMachineException;

	/**
	 * Reverts a previously successful withdrawal from the identifying account.
	 * See {@code #validatePin(int, int)} for the initial checks this method
	 * will do before reverting. Only the last withdraw may be reverted.
	 *
	 * Note, the information about the 'last' withdrawal is stored in the
	 * identifying account. I.e. two people may withdraw from the same account
	 * and both withdrawals may be reverted. Each user may revert their last
	 * withdrawal.
	 *
	 * @returns true if the withdrawal was reverted otherwise false if the given
	 *          tx number was not the last withdrawal
	 * @throws CashMachineException
	 *             if the initial checks do not hold
	 */
	boolean revertWithdraw(int accountId, int pin, int transactionNumber)
	throws CashMachineException;
}
