package cashmachine.api;

import java.util.List;
import java.util.Set;

import cashmachine.model.Bill;

public interface CashMachine
{
	/**
	 * Returns the machine to its initial state.
	 */
	void reset();

	/**
	 * After checking if an account is present and the requested amount can be
	 * divided evenly by 10 and is greater than 0, the amount to be withdrawn is
	 * checked first against the user's account balance and then against the
	 * user's account limit. If the user is able to withdraw a certain amount,
	 * it is deducted from the account and the machine is able to dispense the
	 * money via other methods in this class. If the request is unsuccessful the
	 * machine resets itself.
	 *
	 * A status message is displayed on the screen based on the outcome of this
	 * method.
	 *
	 * @return true if the withdrawal was successful (check against balance and
	 *         limit was successful)
	 * @throws CashMachineException
	 *             if no valid account is present or the amount can not be
	 *             divided evenly by 10.
	 */
	boolean withdraw(int withdrawAmount) throws CashMachineException;

	/**
	 * After checking if an account is present, this method attempts to remove
	 * bank notes from its internal storage. The money is placed in the money
	 * tray and the user is prompted to remove it via the message on the
	 * display.
	 *
	 * The bank notes used can be passed as a parameter, hence they are also
	 * checked against the requested amount.
	 *
	 * @return true if the notes sum up to the withdraw amount and enough bills
	 *         are available
	 * @throws CashMachineException
	 *             if no valid account is present or no withdrawl has happened
	 */
	boolean dispenseAmount(List<Bill> notes) throws CashMachineException;

	/**
	 * After checking if an account is present, this method attempts to remove
	 * bank notes from its internal storage. The money is placed in the money
	 * tray and the user is prompted to remove it via the message on the
	 * display.
	 *
	 * @return true if enough bills are available
	 * @throws CashMachineException
	 *             if no valid account is present or no withdrawl has happened
	 */
	boolean dispenseAmount() throws CashMachineException;

	/**
	 * Sets the account number used for transaction.
	 *
	 * @return true if the account exists and is not banned
	 * @throws CashMachineException
	 *             if there's already an account set
	 */
	boolean setAccountId(int accountId) throws CashMachineException;

	/**
	 * If the account ID which was set is valid and is not banned this method
	 * returns true, otherwise false.
	 */
	boolean isCardValid();

	/**
	 * Checks if the given PIN is valid for the cofigured account. Returns true
	 * if this PIN is valid for the already set account.
	 *
	 * @throws CashMachineException
	 *             if no account number has been set or the account has already
	 *             been validated
	 */
	boolean setPin(int pin) throws CashMachineException;

	/**
	 * Returns true if the set PIN is valid for the currently configured
	 * account, otherwise false.
	 */
	boolean isPinValid();

	/**
	 * Returns a {@code Set} of account numbers, the currently set account user
	 * is allowed to withdraw from.
	 * 
	 * @throws CashMachineException
	 */
	Set<Integer> getWithdrawalAccountIds() throws CashMachineException;

	/**
	 * Sets {@code accountId} as the account to withdraw from.
	 *
	 * @throws CashMachineException
	 *             if the given {@code accountId} isn't a valid withdrawal
	 *             account
	 */
	void setWithdrawalAccountId(int accountId) throws CashMachineException;

	/**
	 * Called after the user has taken the money from the money tray. This sets
	 * the machine to the initial state.
	 */
	void onMoneyTaken();

	/**
	 * If the money tray is open due to calling one of the dispenseAmount
	 * methods, this method returns true, otherwise false
	 */
	boolean isMoneyTrayOpen();

	/**
	 * Returns the message to be shown on the display of the machine
	 */
	String getMessage();

	/**
	 * Returns the amount of 10 Euro bills left in the machine.
	 */
	int getTenFillLevel();

	/**
	 * Returns the amount of 20 Euro bills left in the machine.
	 */
	int getTwentyFillLevel();

	/**
	 * Returns the amount of 50 Euro bills left in the machine.
	 */
	int getFiftyFillLevel();

	/**
	 * Returns the amount of 100 Euro bills left in the machine.
	 */
	int getOneHundredFillLevel();

	/**
	 * Returns the amount of 200 Euro bills left in the machine.
	 */
	int getTwoHundredFillLevel();
}
