package cashmachine.impl;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import cashmachine.api.AccountService;
import cashmachine.api.CashMachine;
import cashmachine.api.CashMachineException;
import cashmachine.api.MachineConnector;
import cashmachine.model.Account;
import cashmachine.model.Bill;

public class CashMachineImpl implements CashMachine
{
	private final MachineConnector machineConnector;
	private final AccountService accountService;
	private int amount10;
	private int amount20;
	private int amount50;
	private int amount100;
	private int amount200;
	private Account currentAccount;
	private String message;
	private int withdrawAmount;
	private int validationTry;
	private boolean moneyTrayOpen;
	private Integer currentAccountId;
	private Integer currentAccountPin;
	private Integer withdrawalAccountId;
	private Account withdrawalAccount;
	private int currentTxNumber;

	public CashMachineImpl(MachineConnector machineConnector, AccountService accountService,
			int amount10, int amount20, int amount50, int amount100, int amount200) {
		this.machineConnector = machineConnector;
		this.accountService = accountService;
		this.amount10 = amount10;
		this.amount20 = amount20;
		this.amount50 = amount50;
		this.amount100 = amount100;
		this.amount200 = amount200;
		init();
	}

	public void reset() {
		if (isPinValid() && withdrawAmount > 0 && currentTxNumber > 0) {
			try {
				accountService.revertWithdraw(currentAccountId, currentAccountPin, currentTxNumber);
			} catch (CashMachineException e) {
				// account was already validated
			}
		}
		init();
	}

	private void init() {
		currentAccountId = null;
		currentAccountPin = null;
		currentAccount = null;
		currentTxNumber = 0;
		withdrawalAccountId = null;
		withdrawalAccount = null;
		withdrawAmount = 0;
		validationTry = 0;
		moneyTrayOpen = false;
		String message = null;
		if (amount10 == 0 && amount20 == 0 && amount50 == 0 && amount100 == 0 && amount200 == 0) {
			message = "Ausser Betrieb, heute ist kein Geld vorhanden!";
		} else if (amount10 > 0 && amount20 > 0 && amount50 > 0 && amount100 > 0 && amount200 > 0) {
			message = "Bankomatkarte einfuehren.";
		} else {
			message = "Bankomatkarte einfuehren. Nur ";
			if (amount200 > 0) {
				message += "100 ";
			}
			if (amount100 > 0) {
				message += "100 ";
			}
			if (amount50 > 0) {
				message += "50 ";
			}
			if (amount20 > 0) {
				message += "20 ";
			}
			if (amount10 > 0) {
				message += "10 ";
			}
			message += "Euro Noten verfuegbar.";
		}
		setMessage(message);
	}

	public boolean withdraw(int withdrawAmount) throws CashMachineException {
		if (!isPinValid() || isMoneyTrayOpen() || withdrawalAccount == null
				|| this.withdrawAmount != 0) {
			throw new CashMachineException();
		}
		if (withdrawAmount < 0 || withdrawAmount % 10 != 0) {
			setMessage("Betrag muss ein Vielfaches sein!");
			return false;
		}
		Account a = withdrawalAccount;
		if (a.getLimit() <= withdrawAmount) {
			setMessage("Limit ueberschritten!");
			return false;
		}
		if (a.getBalance() < withdrawAmount) {
			setMessage("Saldo ueberschritten!");
			return false;
		}
		currentTxNumber = accountService.withdraw(currentAccountId, currentAccountPin,
				withdrawalAccountId,
				withdrawAmount);
		this.withdrawAmount = withdrawAmount;
		return true;
	}

	public boolean dispenseAmount(List<Bill> notes) throws CashMachineException {
		if (!isPinValid() || withdrawAmount == 0 || isMoneyTrayOpen()) {
			throw new CashMachineException();
		}
		int notesTotal = 0;
		int ten = 0, twenty = 0, fifty = 0, onehundred = 0, twohundred = 0;
		for (Bill b : notes) {
			notesTotal += b.getValue();
			switch (b) {
			case TEN:
				ten += 1;
				break;
			case TWENTY:
				twenty += 1;
				break;
			case FIFTY:
				fifty += 1;
				break;
			case ONEHUNDRED:
				onehundred += 1;
				break;
			case TWOHUNDRED:
				twohundred += 2;
				break;
			}
		}
		if (notesTotal != withdrawAmount) {
			setMessage("Auszahlungssumme und Banknotensumme stimmen nicht ueberein!");
			return false;
		}
		if (ten > amount10 || twenty > amount20 || fifty > amount50 || onehundred > amount100
				|| twohundred > amount200) {
			String msg = "Nicht genug Banknoten zur Verfuegung um die Abhebung durchzufuehren!";
			setMessage(msg);
			return false;
			//throw new CashMachineException(msg);
		}
		amount10 -= ten;
		machineConnector.dispenseBill(Bill.TEN, ten);
		amount20 -= twenty;
		machineConnector.dispenseBill(Bill.TWENTY, twenty);
		amount50 -= fifty;
		machineConnector.dispenseBill(Bill.FIFTY, fifty);
		amount100 -= onehundred;
		machineConnector.dispenseBill(Bill.ONEHUNDRED, onehundred);
		amount200 -= twohundred;
		machineConnector.dispenseBill(Bill.ONEHUNDRED, twohundred);
		setMoneyTray(true);
		setMessage("Geld Entnehmen!");
		return true;
	}

	public boolean dispenseAmount() throws CashMachineException {
		if (!isPinValid() || withdrawAmount == 0 || isMoneyTrayOpen()) {
			throw new CashMachineException();
		}
		final int rollbackAmount = withdrawAmount;
		int dispensed = 0;
		int ten = 0, twenty = 0, fifty = 0, onehundred = 0, twohundred = 0;
		while (withdrawAmount > 500) {
			if ((dispensed = dispenseTwoHundred(2)) > 0) {
				withdrawAmount -= dispensed * 200;
				twohundred += dispensed;
			} else {
				break;
			}
		}
		while (withdrawAmount > 200) {
			boolean changed = false;
			if ((dispensed = dispenseOneHundred(1)) > 0) {
				withdrawAmount -= dispensed * 100;
				onehundred += dispensed;
				changed = true;
			}
			if ((dispensed = dispenseTwenty(4)) > 0) {
				withdrawAmount -= dispensed * 20;
				twenty += dispensed;
				changed = true;
			}
			if (!changed) {
				break;
			}
		}
		while (withdrawAmount > 110) {
			boolean changed = false;
			if ((dispensed = dispenseFifty(1)) > 0) {
				withdrawAmount -= dispensed * 50;
				fifty += dispensed;
				changed = true;
			}
			if ((dispensed = dispenseTwenty(3)) > 0) {
				withdrawAmount -= dispensed * 20;
				twenty += dispensed;
				changed = true;
			}
			if (!changed) {
				break;
			}
		}
		while (withdrawAmount > 50) {
			if ((dispensed = dispenseFifty(1)) > 0) {
				withdrawAmount -= dispensed * 50;
				fifty += dispensed;
			} else {
				break;
			}
		}
		while (withdrawAmount > 0) {
			if ((dispensed = dispenseTen(1)) > 0) {
				withdrawAmount -= dispensed * 10;
				ten += dispensed;
			} else {
				break;
			}
		}
		if (withdrawAmount > 0) {
			withdrawAmount = rollbackAmount;
			amount10 += ten;
			amount20 += twenty;
			amount50 += fifty;
			amount100 += onehundred;
			amount200 += twohundred;
			String msg = "Nicht genug Banknoten zur Verfuegung um die Abhebung durchzufuehren!";
			setMessage(msg);
			return false;
		}
		machineConnector.dispenseBill(Bill.TEN, ten);
		machineConnector.dispenseBill(Bill.TWENTY, twenty);
		machineConnector.dispenseBill(Bill.FIFTY, fifty);
		machineConnector.dispenseBill(Bill.ONEHUNDRED, onehundred);
		machineConnector.dispenseBill(Bill.TWOHUNDRED, twohundred);
		setMoneyTray(true);
		setMessage("Geld Entnehmen!");
		return true;
	}

	private int dispenseTen(int amount) {
		if (amount10 > 0) {
			int d = Math.min(amount, amount10);
			amount10 -= d;
			return d;
		}
		return 0;
	}

	private int dispenseFifty(int amount) {
		if (amount50 > 0) {
			int d = Math.min(amount, amount50);
			amount50 -= d;
			return d;
		}
		return 0;
	}

	private int dispenseTwenty(int amount) {
		if (amount20 > 0) {
			int d = Math.min(amount, amount20);
			amount20 -= d;
			return d;
		}
		return 0;
	}

	private int dispenseOneHundred(int amount) {
		if (amount100 > 0) {
			int d = Math.min(amount, amount100);
			amount100 -= d;
			return d;
		}
		return 0;
	}

	private int dispenseTwoHundred(int amount) {
		if (amount100 > 0) {
			int d = Math.min(amount, amount100);
			amount100 -= d;
			return d;
		}
		return 0;
	}

	public boolean setAccountId(int accountId) throws CashMachineException {
		if (currentAccountId != null) {
			throw new CashMachineException();
		}
		if (accountService.isValidAccountId(accountId)) {
			currentAccountId = accountId;
		}
		if (isCardValid()) {
			setMessage("PIN eingeben:");
		}
		return isCardValid();
	}

	public boolean isCardValid() {
		return currentAccountId != null;
	}

	public boolean setPin(int pin) throws CashMachineException {
		if (!isCardValid() || isPinValid() || withdrawAmount > 0 || isMoneyTrayOpen()) {
			throw new CashMachineException();
		}
		++validationTry;
		currentAccount = accountService.validatePin(currentAccountId, pin);
		if (validationTry == 4 && currentAccount == null) {
			accountService.banAccount(currentAccountId);
			reset();
			return false;
		}
		if (isPinValid()) {
			currentAccountPin = pin;
			setMessage("Waehlen Sie Abbuchungskonto.");
		}
		return isPinValid();
	}

	public boolean isPinValid() {
		return currentAccount != null;
	}

	public Set<Integer> getWithdrawalAccountIds() throws CashMachineException {
		if (!isCardValid() || !isPinValid() || withdrawAmount > 0 || isMoneyTrayOpen()) {
			throw new CashMachineException();
		}
		Set<Integer> ids = new HashSet<Integer>(currentAccount
				.getAuthorizedAccountIds());
		ids.add(currentAccount.getAccountId());
		return ids;
	}

	public void setWithdrawalAccountId(int accountId) throws CashMachineException {
		if (!isCardValid() || !isPinValid() || withdrawAmount > 0 || isMoneyTrayOpen()) {
			throw new CashMachineException();
		}
		if (accountId != currentAccount.getAccountId()
				&& !currentAccount.isAuthorizedToSign(accountId)) {
			throw new CashMachineException();
		}
		withdrawalAccountId = accountId;
		withdrawalAccount = accountService.getWithdrawalAccount(currentAccountId,
				currentAccountPin, withdrawalAccountId);
		setMessage("Waehlen Sie Betrag.");
	}

	public void onMoneyTaken() {
		init();
	}

	public boolean isMoneyTrayOpen() {
		return moneyTrayOpen;
	}

	private void setMoneyTray(boolean state) throws CashMachineException {
		moneyTrayOpen = state;
		if (state) {
			machineConnector.openMoneyTray();
		}
	}

	public String getMessage() {
		return message;
	}

	private void setMessage(String message) {
		this.message = message;
	}

	public int getTenFillLevel() {
		return amount10;
	}

	public int getTwentyFillLevel() {
		return amount20;
	}

	public int getFiftyFillLevel() {
		return amount50;
	}

	public int getOneHundredFillLevel() {
		return amount100;
	}

	public int getTwoHundredFillLevel() {
		return amount200;
	}
}
