package cashmachine;

import static org.hamcrest.Matchers.is;
import static org.junit.Assert.assertThat;
import static org.mockito.Matchers.anyInt;
import static org.mockito.Matchers.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import java.util.Arrays;

import org.junit.Test;

import cashmachine.api.AccountService;
import cashmachine.api.CashMachine;
import cashmachine.api.CashMachineException;
import cashmachine.api.MachineConnector;
import cashmachine.impl.CashMachineImpl;
import cashmachine.model.Account;

public class CashMachineTest
{
	private static final int AC1_ID = 1;
	private static final int AC1_PIN = 10;
	private static final int AC1_BALANCE = 10000;
	private static final int AC1_LIMIT = 1000;
	private static final int AC2_ID = 2;
	private static final int AC2_PIN = 20;
	private static final int AC2_BALANCE = 22;
	private static final int AC2_LIMIT = 1000;
	private static final int INVALID_ID = 3;
	private static final int DEFAULT_FILL_10 = 100;
	private static final int DEFAULT_FILL_20 = 50;
	private static final int DEFAULT_FILL_50 = 20;
	private static final int DEFAULT_FILL_100 = 10;
	private static final int DEFAULT_FILL_200 = 5;

	private Account createAccount1() {
		return new Account(AC1_ID, AC1_PIN, AC1_BALANCE, AC1_LIMIT, Arrays.asList(AC2_ID));
	}

	private Account createAccount2() {
		return new Account(AC2_ID, AC2_PIN, AC2_BALANCE, AC2_LIMIT, Arrays.<Integer> asList());
	}

	private MachineConnector createMockedConnector() {
		return mock(MachineConnector.class);
	}

	private AccountService createMockedAccountService() throws CashMachineException {
		AccountService a = mock(AccountService.class);
		when(a.isValidAccountId(AC1_ID)).thenReturn(true);
		when(a.validatePin(AC1_ID, AC1_PIN)).thenReturn(createAccount1());
		when(a.getWithdrawalAccount(AC1_ID, AC1_PIN, AC1_ID)).thenReturn(createAccount1());
		when(a.getWithdrawalAccount(AC1_ID, AC1_PIN, AC2_ID)).thenReturn(createAccount2());
		when(a.withdraw(eq(AC1_ID), eq(AC1_PIN), eq(AC1_PIN), anyInt())).thenReturn(1);
		when(a.withdraw(eq(AC1_ID), eq(AC1_PIN), eq(AC2_PIN), anyInt())).thenReturn(1);
		return a;
	}

	private CashMachine createCashMachine() throws CashMachineException {
		return createCashMachine(createMockedConnector());
	}

	private CashMachine createCashMachine(MachineConnector machineConnector)
	throws CashMachineException {
		return createCashMachine(machineConnector, createMockedAccountService());
	}

	private CashMachine createCashMachine(AccountService accountService) {
		return createCashMachine(createMockedConnector(), accountService);
	}

	private CashMachine createCashMachine(MachineConnector connector, AccountService accountService) {
		return createCashMachine(connector, accountService, DEFAULT_FILL_10,
				DEFAULT_FILL_20, DEFAULT_FILL_50, DEFAULT_FILL_100, DEFAULT_FILL_200);
	}

	private CashMachine createCashMachine(MachineConnector machineConnector,
			AccountService accountService, int amount10, int amount20, int amount50, int amount100,
			int amount200) {
		return new CashMachineImpl(machineConnector, accountService, amount10, amount20, amount50,
				amount100, amount200);
	}

	@Test
	public void fillLevels_shouldReturnRightAmount() throws CashMachineException {
		CashMachine cm = createCashMachine();
		assertThat(cm.getTenFillLevel(), is(DEFAULT_FILL_10));
		assertThat(cm.getTwentyFillLevel(), is(DEFAULT_FILL_20));
		assertThat(cm.getFiftyFillLevel(), is(DEFAULT_FILL_50));
		assertThat(cm.getOneHundredFillLevel(), is(DEFAULT_FILL_100));
		assertThat(cm.getTwoHundredFillLevel(), is(DEFAULT_FILL_200));
	}

	// TODO put your tests here
}
